<?php

namespace Phake\Annotation;

use ReflectionClass;
use ReflectionProperty;

/*
 * Phake - Mocking Framework
 *
 * Copyright (c) 2010-2021, Mike Lively <mike.lively@sellingsource.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  *  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  *  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *
 *  *  Neither the name of Mike Lively nor the names of his
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * @category   Testing
 * @package    Phake
 * @author     Mike Lively <m@digitalsandwich.com>
 * @copyright  2010 Mike Lively <m@digitalsandwich.com>
 * @license    http://www.opensource.org/licenses/bsd-license.php  BSD License
 * @link       http://www.digitalsandwich.com/
 */

class LegacyReader implements IReader
{
    private const ANNOTATION_NAME = 'Mock';

    public function getPropertiesWithMockAnnotation(ReflectionClass $class): iterable
    {
        $properties = [];
        foreach ($class->getProperties() as $property) {
            $annotations = $this->getPropertyAnnotations($property);

            if (isset($annotations[self::ANNOTATION_NAME])) {
                $properties[] = $property;
            }
        }

        return $properties;
    }

    public function getMockType(ReflectionProperty $property): ?string
    {
        $mockedClass = null;
        $annotations = $this->getPropertyAnnotations($property);

		if ($annotations['Mock'] !== true) {
			$mockedClass = $annotations['Mock'];
		} elseif (isset($annotations['var'])) {
			$mockedClass = $annotations['var'];
		}

		if ($mockedClass && substr($mockedClass, 0, 1) !== '\\' &&  $this->useDoctrineParser()) {
			$parser = new \Doctrine\Common\Annotations\PhpParser();

            $reflectionClass = $property->getDeclaringClass();
            $useStatements   = $parser->parseClass($reflectionClass);
            $key             = strtolower($mockedClass);

            if (array_key_exists($key, $useStatements)) {
                $mockedClass = $useStatements[$key];
            } elseif ($reflectionClass->getNamespaceName() && $this->definedUnderTestNamespace($mockedClass, $reflectionClass->getNamespaceName())) {
                $mockedClass = $reflectionClass->getNamespaceName() . '\\' . $mockedClass;
            }
		}

        return $mockedClass;
    }

    /**
     * Returns an associative array containing a property's annotations and their values.
     *
     * @param string $property
     *
     * @return array
     */
    private function getPropertyAnnotations(\ReflectionProperty $property)
    {
        $comment = $property->getDocComment();

        $annotations = array();
        foreach (explode("\n", $comment) as $line) {
            if (preg_match('#^\s+\*\s*@(\w+)(?:\s+(.*))?\s*$#', $line, $matches)) {
                $annotations[$matches[1]] = isset($matches[2]) ? $matches[2] : true;
            }
        }
        return $annotations;
    }

    private function useDoctrineParser(): bool
    {
        return class_exists('Doctrine\Common\Annotations\PhpParser');
    }

    private function definedUnderTestNamespace($mockedClass, $testNamespace): bool
    {
        return class_exists($testNamespace . '\\' . $mockedClass);
    }
}
